#!/usr/bin/perl -w

# Try to extract the content of a MIME message as plain text

# This code can be used under the same terms as Perl itself. It comes
# with absolutely NO WARRANTY. Use at your own risk.

# Andreas Amann, 2002-2004
# http://homepage.mac.com/aamann/

# since we use some modules not installed in OSX per default, we have them packaged within the app
# -> add the local path to the include path
use lib 'Perl';

#use strict;
use warnings;
use MIME::Parser;
use HTML::Parser 3.00 ();
use MIME::WordDecoder;
use Unicode::Map8;
use Getopt::Std;
use Text::Wrap;
use Encode 'from_to';

use vars qw($wd $e $x $isQP $isUTF8 $quoteString);
my %inside;
my %options=();
$quoteString = '';

getopts("af:w:",\%options);
die "Usage: ExtractText -f filename [-w number of columns to wrap text at]\n" unless defined $options{f};
if ( $options{w} ) {
	$Text::Wrap::columns = $options{w};
	$Text::Wrap::huge = 'overflow';
}

#Setup our charset maps
my $mac_map = Unicode::Map8->new("MacRoman") || die;
my $latin1_map = Unicode::Map8->new("latin1") || die;

#Setup the MIME parser
my $mime_parser = new MIME::Parser;
$mime_parser->ignore_errors(1);
$mime_parser->extract_uuencode(1);
$mime_parser->output_dir("/tmp");	# so we don't litter the file system in case something goes wrong

# Setup the HTML parser
my $html_parser = HTML::Parser->new
(
 api_version => 3,
 handlers    =>
 [
	 start => [\&tag, "tagname, '+1',tag,attr"],
	 end   => [\&tag, "tagname, '-1',tag,attr"],
	 text  => [\&text, "dtext" ],
	 ],
 marked_sections => 1
 ) or die "Can't create HTML parser\n";

$html_parser->ignore_elements(qw(title style)); # some clients seem to add the subject as the title -> ignore it
												# since we convert to plain text, we don't care about styles either

# Setup the text/enriched parser - we use a slightly simpler HTML parser for that
my $richtext_parser = HTML::Parser->new
(
 api_version => 3,
 handlers    =>
 [
	 start => [\&tag, "tagname, '+1',tag,attr"],
	 end   => [\&tag, "tagname, '-1',tag,attr"],
	 text  => [\&richtext, "dtext" ],
	 ],
 marked_sections => 1
 ) or die "Can't create richtext parser\n";

$richtext_parser->ignore_elements(qw(param)); # since we don't care about the tags, we also can ignore their parameters

# Pass the message through the MIME parser
eval { $e = $mime_parser->parse_open($options{f}); };
my $error = ($@ || $mime_parser->last_error);
if ($error)
{
	# Get rid of the temp files
	$mime_parser->filer->purge;
	# and pass on the error
	die "Error parsing the message: $error\n";
}
# remove duplicates caused by multipart/alternative
$e = &analyze($e);

# Setup the decoder based on the header information
setup_decoder($e->head);
# print the headers as needed
if ( $options{a} ) {
	print mac_decode($e->head->as_string);
}
else {
	print_header("From");
	print_header("Subject");
	print_header("Date");
	print_header("To");
	print_header("Cc");
	print_header("Bcc");
}
# finish the header
print "\n";
# and process each of the parts
decode_entities($e);

# all done -> delete temp files
$mime_parser->filer->purge;

####################################

# tag handler for both, HTML and rich text processing - produce something rather close to the HTML version
sub tag
{
	my($tagname, $num, $tag, $attr) = @_;
	my $skip = 0;
	$inside{$tagname} += $num;
	if ( $tag eq "br" || $tag eq "div" ) {
		print "\n";											# start a new line
	}
	elsif ( $tag eq "blockquote" || $tag eq "excerpt" ) {
		print "\n";											# start a new line
		$quoteString .= "> ";
	}
	elsif ( $tag eq "/blockquote" || $tag eq "/excerpt" ) {
		print "\n";											# start a new line
			$quoteString = substr($quoteString, 0, length($quoteString) - 2 );
	}
	elsif ( $tag eq "/p" ) {
		print "\n\n";										# start a new paragraph (i.e., insert empty line)
	}
	elsif ( $tag eq "hr" ) {
		print "\n----------\n";								# create a dummy <HR>
	}
	elsif ( $tag eq "a" ) {
		return unless exists $attr->{href};
		print "<$attr->{href}> ";							# insert the link location
	}
	elsif ( $tag eq "img" ) {
		if ( exists $attr->{alt} ) {
			print "[Image '$attr->{alt}' <$attr->{src}>]";	# insert IMG placeholder with title ("alt") and original location
		}
		else {
			print "[Image <$attr->{src}>] ";				# insert IMG placeholder with original location only
		}
	}
	else {
		$skip = 1;
	}
	if ( $skip == 0 && $inside{blockquote} ) {
		print $quoteString;
	}
}

# text handler for the text extracted from the HTML source
sub text
{
	# we don't need anything within a script tag
	return if $inside{script};
	# pass through the character decoder and convert ISO-8859-1 to Mac encoding if needed
	$_ = complete_decode($_[0]);
	# remove leading newline (which would have been inserted if we had a tag as the last thing in the line before)
	s/^\n//g;
	# and add a space at the end unless we have an empty line
	s/(.)\n$/$1 /g;
	# replace all multiple white-space with a single space (HTML does so)
	s/\s+/ /g;
	# if the -w flag was supplied, wrap long lines as requested
	if ( $options{w} ) { $_ = wrap( '', $quoteString, $_ ); }
	# print the text after all processing
	print;
}

# text handler for the text extracted from the rich text source
sub richtext
{
	# pass through the character decoder and convert ISO-8859-1 to Mac encoding if needed
	$_ = complete_decode($_[0]);
	# remove leading newline (which would have been inserted if we had a tag as the last thing in the line before)
	s/^\n$/ /g;
	s/^\n//g;
	# and add a space at the end unless we have a real newline
	s/([^\n])\n$/$1 /g;
	# replace single newlines with spaces
	s/([^\n])\n([^\n])/$1 $2/g;
	# remove the additional newline from "real" newlines
	s/\n(\n+)/$1/g;
	# if the -w flag was supplied, wrap long lines as requested
	if ( $options{w} ) { $_ = wrap( '', $quoteString, $_ ); }
	# print the text after all processing
	print;
}

# set up decoder based on header information for the current MIME entity
sub setup_decoder
{
	my $head = shift;
	$isUTF8 = 0;
	if ( $head->get('Content-Type') && $head->get('Content-Type') =~ m!charset="([^\"]+)"! ) {
		$wd = supported MIME::WordDecoder uc $1;
		$isUTF8 = ( $1 =~ m/utf-?8/i );
	}
	$wd = supported MIME::WordDecoder "ISO-8859-1" unless $wd;
	$isQP = ( $head->get('Content-Transfer-Encoding') && $head->get('Content-Transfer-Encoding') =~ /quoted-printable/i );
}

# decode each of the MIME entities we care about
sub decode_entities
{
	my $ent = shift;
	if ( my @parts = $ent->parts ) {
		decode_entities($_) for @parts;
	}
	elsif ( my $body = $ent->bodyhandle ) {
		my $type = $ent->head->mime_type;
		setup_decoder($ent->head);
		if ( $type =~ /text\/plain/i ) {	# dump the content directly (after passing through the WordDecoder)
			my $textpart = complete_decode($body->as_string);
			if ( $options{w} ) {
			  	@lines = split(/\n/, $textpart);
				foreach $line (@lines) {
					if ( $line =~ m/^(>[> ]+)/ ) {
						print wrap('',$1,$line),"\n";
					}
					else {
						print wrap('','',$line),"\n";
					}
				}
			}
			else {
				print $textpart;
			}
		}
		elsif ( $type =~ /text\/html/i ) {
			$html_parser->parse($body->as_string);		# pass the content on to the HTML parser which creates the output
		}
		elsif ( $type =~ /text\/enriched/i ) {
			$richtext_parser->parse($body->as_string);	# pass the content on to the richtext parser which creates the output
		}
		elsif ( $type =~ /application\/applefile/i ) {
			# nothing here in order not to get duplicate attachment listings for files with resource and data fork
		}
		else {
			if ( my $filename = $ent->head->recommended_filename() ) {
				print "\n[Attachment '", mac_decode($filename),"']";		# add attachment placeholder with original filename
			}
			else {
				print "\n[Unhandled part of type $type]";# don't know how to deal with this -> add unspecified placeholder
			}
		}
	}
}

# get rid of MIME entities we don't care about before processing
sub analyze {
	my($body,$i)=(@_,0);
	my($parts)=$body->{ME_Parts};
	if ( $body->mime_type =~ /multipart\/alternative/i ) {
		# Reduce multipart/alternative
		$i=-1;
		$i++ until ( $$parts[$i]->mime_type =~ /text\/plain/i || $i==$#{$parts} );
		if ( $$parts[$i]->mime_type =~ /text\/plain/i ) {
			@$parts = @$parts[$i];
		}
	}
	else {
		# Kill superfluous junk:
		# - text/x-vcard
		# - application/pgp-signature
		# - application/x-pkcs7-signature and application/pkcs7-signature
		# - application/ms-tnef
		# Recursion on multipart
		while ( $i <= $#{$parts} ) {
			if ( $$parts[$i]->mime_type =~ /(text\/x-vcard|application\/(?:(?:pgp|(?:x-)?pkcs7)-signature|ms-tnef))/i ) {
				splice(@$parts,$i,1);
			}
			elsif ( $$parts[$i]->mime_type =~ /^multipart\//i ) {
				   $$parts[$i] = &analyze($$parts[$i]);
				   $i++;
			}
			else {
				$i++;
			}
		}
	}
	$body->{ME_Parts} = $parts;

	if ( $body->mime_type =~ /^multipart\/related/i && $body->head->mime_attr("content-type.type") =~ m/multipart\/alternative/i ) {
		$body->head->mime_attr("content-type.type" => "text/plain");
	}
	$body->make_singlepart if $body->parts == 1;
	return $body;
}

sub complete_decode {
	my $encoded_text = shift;
	if ( $isQP || $isUTF8 ) {
		return mac_decode($encoded_text);
	}
	else {
		return $wd->decode($encoded_text);
	}
}

sub mac_decode {
	my $decoded_string = $wd->decode(shift);
	if ( $isUTF8) {
	    from_to($decoded_string,"utf-8","latin1");
	}
	$decoded_string = $latin1_map->recode8($mac_map, $decoded_string);
	return $decoded_string;
}

sub print_header {
	my $the_header = shift;
	if ( $e->head->get($the_header) ) { print $the_header, ": ", mac_decode($e->head->get($the_header)); }
}

# EOF
